/*
* Copyright 2016 The CHOReVOLUTION project
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
*      http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*/
package eu.chorevolution.transformations.generativeapproach.cdgenerator.util;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;

import org.apache.commons.io.FileUtils;
import org.apache.commons.lang3.StringUtils;
import org.apache.ws.commons.schema.XmlSchema;
import org.eclipse.bpmn2.Choreography;
import org.eclipse.bpmn2.ChoreographyActivity;
import org.eclipse.bpmn2.ChoreographyTask;
import org.eclipse.bpmn2.Definitions;
import org.eclipse.bpmn2.DocumentRoot;
import org.eclipse.bpmn2.EndEvent;
import org.eclipse.bpmn2.ExclusiveGateway;
import org.eclipse.bpmn2.ExtensionAttributeValue;
import org.eclipse.bpmn2.FlowElement;
import org.eclipse.bpmn2.FlowNode;
import org.eclipse.bpmn2.FormalExpression;
import org.eclipse.bpmn2.Gateway;
import org.eclipse.bpmn2.GatewayDirection;
import org.eclipse.bpmn2.Import;
import org.eclipse.bpmn2.Message;
import org.eclipse.bpmn2.MessageFlow;
import org.eclipse.bpmn2.Participant;
import org.eclipse.bpmn2.SequenceFlow;
import org.eclipse.bpmn2.StartEvent;
import org.eclipse.bpmn2.SubChoreography;
import org.eclipse.bpmn2.impl.FormalExpressionImpl;
import org.eclipse.bpmn2.util.Bpmn2ResourceFactoryImpl;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.eclipse.emf.ecore.util.FeatureMap;
import org.eclipse.emf.ecore.xml.type.impl.AnyTypeImpl;
import org.slf4j.ext.XLogger;
import org.slf4j.ext.XLoggerFactory;

import eu.chorevolution.transformations.generativeapproach.cdgenerator.CDGeneratorException;
import eu.chorevolution.transformations.generativeapproach.cdgenerator.model.ChoreographyData;
import eu.chorevolution.transformations.generativeapproach.cdgenerator.model.ChoreographyMessageInformation;
import eu.chorevolution.transformations.generativeapproach.cdgenerator.model.ProjectionData;
import eu.chorevolution.transformations.generativeapproach.cdgenerator.model.TaskCorrelationData;


public class BPMNUtils {
	
	private static final String BPMN_EXTENSION = ".bpmn";  
	private static final String EPROXY_URI_MESSAGE_ITEM_REF_STRUCTURE_REF_MESSAGE_TYPE_NAME_SEPARATOR = "#";
	private static final List<String> XPATH_OPERATORS = new ArrayList<String>(Arrays
			.asList("(",")","|","+","-","*",
			"div","=","!=","<","<=",">",">=","or","and","mod","count","count-not-empty","id","concat",
			"substring-before","substring-after","substring","normalize-space","translate","lower-case",
			"upper-case","ends-with","boolean","not","sum","floor","ceiling","round","avg","min","max"));
	
	private static final XLogger LOGGER = XLoggerFactory.getXLogger(BPMNUtils.class);	
	
	public static ChoreographyData readChoreographyData(ProjectionData projectionData)
			throws CDGeneratorException {
		
		// log execution flow
		LOGGER.entry("CD name: "+projectionData.getCdName()," CD Participant name: "
				+projectionData.getParticipantName());
		
		// load and extract choreography contained in the BPMN model.
		String tempFolderPath = Utils.createTemporaryFolderFromMillisAndGetPath();
		File bpmnFile = new File(tempFolderPath+File.separator+projectionData.getCdName()+BPMN_EXTENSION);
		ChoreographyData choreographyData = null;
		try {
			FileUtils.writeByteArrayToFile(bpmnFile, projectionData.getBpmn());
			// load and extract choreographies contained in the BPMN model.
			List<Choreography> choreographies = BPMNUtils.loadChoreographiesFromBPMN2Model(bpmnFile);					
			XmlSchema schema = XSDSchemaUtils.getXmlSchema(projectionData.getTypes());
			choreographyData = new ChoreographyData(choreographies, schema);
		} catch (IOException e) {
			LOGGER.error(e.getMessage(),e);
			throw new CDGeneratorException("Exception in readChoreographyData see log file for details");
		}
						
		Utils.deleteFolder(tempFolderPath);
		
		// log execution flow
		LOGGER.exit();
		
		return choreographyData;
	}
	
	public static ChoreographyData readChoreographyData(File bpmnFile) throws CDGeneratorException {
			
		// log execution flow
		LOGGER.entry();
		
		// load and extract choreographies contained in the BPMN model.
		List<Choreography> choreographies = BPMNUtils.loadChoreographiesFromBPMN2Model(bpmnFile);
		
		XmlSchema schema = null;
		// get the container from the first choreography in order to retrive the imports
		List<Import> imports = ((Definitions) choreographies.get(0).eContainer()).getImports();
		
		if (!imports.isEmpty()) {
			// TODO for now we assume that there is only one import
			try {
				schema = XSDSchemaUtils.getXmlSchema(FileUtils.readFileToByteArray(new File(bpmnFile
						.getParent(), imports.get(0).getLocation())));
			} catch (IOException e) {
				LOGGER.error(e.getMessage(),e);
				throw new CDGeneratorException("Exception in readChoreographyData see log file for details");
			}
		}
		ChoreographyData choreographyData = new ChoreographyData(choreographies, schema);

		// log execution flow
		LOGGER.exit();
		
		return choreographyData;
	}	

	public static Choreography getMainChoreographyFromChoreographyData(ChoreographyData choreographyData){
		
		// log execution flow
		LOGGER.entry();
		for (Choreography choreographyItem1 : choreographyData.getChoreographies()) {
			// check if choreographyItem1 is referenced inside the others choreographies
			boolean isChoregraphyReferenced = false;
			for (Choreography choreographyItem2 : choreographyData.getChoreographies()) {
				if(choreographyItem2 != choreographyItem1 && 
				   BPMNUtils.isChoreographyRefencedInsideChoreography(choreographyItem2,
						   choreographyItem1.getName())){
					isChoregraphyReferenced = true;
				}			
			}
			// choreographyItem1 is not referenced in all the others choreographies 
			if(!isChoregraphyReferenced)
				// log execution flow
				LOGGER.exit();
				return choreographyItem1;
		}
		
		throw new CDGeneratorException("No main choreography found see log file for details.");
	}
	
	public static Choreography getMainChoreographyFromChoreographyies(List<Choreography> choreographies) {
		
		// log execution flow
		LOGGER.entry();
		for (Choreography choreographyItem1 : choreographies) {
			// check if choreographyItem1 is referenced inside the others choreographies 
			boolean isChoregraphyReferenced = false;
			for (Choreography choreographyItem2 : choreographies) {
				if(choreographyItem2 != choreographyItem1 && 
				   BPMNUtils.isChoreographyRefencedInsideChoreography(choreographyItem2,
						   choreographyItem1.getName())){
					isChoregraphyReferenced = true;
				}			
			}
			// choreographyItem1 is not referenced in all the others choreographies 
			if(!isChoregraphyReferenced)
				// log execution flow
				LOGGER.exit();
				return choreographyItem1;
		}
		
		throw new CDGeneratorException("No main choreography found see log file for details.");		
	}
	
	public static boolean isChoreographyRefencedInsideChoreography(Choreography choreography,
			String choreographyName){
		
		// log execution flow
		LOGGER.entry();
		// iterate over every sub-choregraphy
		// to check is the choregraphy with choreographyName
		// is referenced as a sub-choreography
		for (FlowElement flowElement : choreography.getFlowElements()) {
			// flowNode is a sub-choreography
			if(flowElement instanceof SubChoreography){
				if(((SubChoreography)flowElement).getName().equals(choreographyName)){
					// log execution flow
					LOGGER.exit();
					return true;
				}	
			}
		}
		// log execution flow
		LOGGER.exit();
		return false;
	}
	
	public static Choreography getChoreography(List<Choreography> choreographies, String choreographyName){
		
		// log execution flow
		LOGGER.entry();
		for (Choreography choreography : choreographies) {
			if(choreography.getName().equals(choreographyName)){
				LOGGER.info("Choreography: "+choreographyName+" found!");			
				// log execution flow
				LOGGER.exit();				
				return choreography;
			}	
		}
		LOGGER.info("NO Choreography: "+choreographyName+" found!");
		throw new CDGeneratorException("NO Choreography: "+choreographyName
				+" found see log file for details.");
	}
	
	public static Participant getParticipantFromName(ChoreographyData choreographyData, String name)
			throws CDGeneratorException {

		// log execution flow
		LOGGER.entry();
		for (Choreography choreography : choreographyData.getChoreographies()) {
			for (Participant participant : choreography.getParticipants()) {
				if (participant.getName().equals(name)){
					LOGGER.info("Participant: " +name+" found.");
					// log execution flow
					LOGGER.exit();	
					return participant;
				}	
			}			
		}
		LOGGER.info("Participant: " +name+" not found.");
		throw new CDGeneratorException("Participant: " +name+" not found see log file for details.");
	}

	public static Participant getParticipantFromName(Choreography choreography, String name)
			throws CDGeneratorException {

		// log execution flow
		LOGGER.entry();
		for (Participant participant : choreography.getParticipants()) {
			if (participant.getName().equals(name)){
				LOGGER.info("Participant "+name+" found inside choreography "+choreography.getName());
				// log execution flow
				LOGGER.exit();				
				return participant;
			}	
		}
		LOGGER.info("Participant "+name+" not found inside choreography "+choreography.getName());
		throw new CDGeneratorException("Participant "+name+" not found inside choreography "
				+choreography.getName()+" see log file for details.");
	}

	public static Participant getParticipantFromName(List<Choreography> choreographies, String name)
			throws CDGeneratorException {

		// log execution flow
		LOGGER.entry();
		for (Choreography choreography : choreographies) {
			for (Participant participant : choreography.getParticipants()) {
				if (participant.getName().equals(name)){
					LOGGER.info("Participant: "+name+" found.");				
					// log execution flow
					LOGGER.exit();
					return participant;
				}	
			}			
		}
		LOGGER.info("Participant: "+name+" not found.");
		throw new CDGeneratorException("Participant: "+name+" not found see log file for details.");
	}	
	
	public static Participant getInitiatingParticipant(ChoreographyActivity choreographyActivity)
			throws CDGeneratorException {
		
		// log execution flow
		LOGGER.entry();
		if (choreographyActivity.getInitiatingParticipantRef() != null){
			LOGGER.info("Initiating participant found in the choreography activity: "
					+ choreographyActivity.getName());	
			// log execution flow
			LOGGER.exit();			
			return choreographyActivity.getInitiatingParticipantRef();
		}
		LOGGER.info("No Initiating participant found in the choreography activity: "
				+ choreographyActivity.getName());
		throw new CDGeneratorException("No Initiating participant found in the choreography activity: " 
				+ choreographyActivity.getName()+" not found see log file for details.");
	}

	public static Participant getReceivingParticipant(ChoreographyActivity choreographyActivity)
			throws CDGeneratorException {
		
		// log execution flow
		LOGGER.entry();			
		for (Participant participant : choreographyActivity.getParticipantRefs()) {
			if (!participant.equals(choreographyActivity.getInitiatingParticipantRef())) {
				LOGGER.info("Receiving participant found in the choreography activity: "
						+ choreographyActivity.getName());			
				// log execution flow
				LOGGER.exit();				
				return participant;
			}
		}
		LOGGER.info("No Receiving participant found in the choreography activity: "
				+ choreographyActivity.getName());
		throw new CDGeneratorException("No Receiving participant found in the choreography activity: " 
				+ choreographyActivity.getName()+" not found see log file for details.");
	}

	
    public static StartEvent getStartEvent(Choreography choreography) throws CDGeneratorException {
    	
		// log execution flow
		LOGGER.entry();	
        // we assume that the choreography has only one start event
        for (FlowElement flowElement : choreography.getFlowElements()) {
            if (flowElement instanceof StartEvent) {
				LOGGER.info("Start event found in the BPMN2 Choreography: " + choreography.getName());			
				// log execution flow
				LOGGER.exit();         	
                return (StartEvent) flowElement;
            }
        }
		LOGGER.info("No start event found in the BPMN2 Choreography: " + choreography.getName());	        
        throw new CDGeneratorException("No start event found in the BPMN2 Choreography: " 
        		+ choreography.getName() +" not found see log file for details.");
   }
    
   public static String getFirstChoregraphyTaskName(Choreography choreography){
	   
		// log execution flow
		LOGGER.entry();	
		String firstChoreographyTaskName = ((ChoreographyTask)BPMNUtils.getStartEvent(choreography)
				.getOutgoing().get(0).getTargetRef()).getName();
		// we assume that a start has only one outgoing, whose target is a choreography task
		LOGGER.info("First choregraphy task name: " + firstChoreographyTaskName);	
		return firstChoreographyTaskName;
   }
	
   public static String getMessageNameChoreographyTaskSentFromParticipant(ChoreographyTask choreographyTask, 
		   Participant participant){

	   // log execution flow
	   LOGGER.entry();		   
	   for (MessageFlow messageFlow : choreographyTask.getMessageFlowRef()) {
		   if(messageFlow.getSourceRef().equals(participant)){
			   LOGGER.info("The message sent from participant: "+participant.getName()
			   		+" within choreography task "+choreographyTask.getName()+" is "
					+messageFlow.getMessageRef().getName());			   
			   // log execution flow
			   LOGGER.exit();			   
			   return messageFlow.getMessageRef().getName();
		   }
	   }		
	   LOGGER.info("No message sent from participant: "+participant.getName()+" within choreography task "
			   +choreographyTask.getName()+" found.");	        
	   // log execution flow
	   LOGGER.exit();
	   return null;
   }
 
   public static String getMessageNameChoreographyTaskSentToParticipant(ChoreographyTask choreographyTask, 
		   Participant participant){

	   // log execution flow
	   LOGGER.entry();
	   for (MessageFlow messageFlow : choreographyTask.getMessageFlowRef()) {
		   if(messageFlow.getTargetRef().equals(participant)){
			   LOGGER.info("The message sent to participant: "+participant.getName()
			   		   +" within choreography task "+choreographyTask.getName()+" is "
					   +messageFlow.getMessageRef().getName());				   
			   // log execution flow
			   LOGGER.exit();	
			   return messageFlow.getMessageRef().getName();
		   }
	   }		
	   LOGGER.info("No message sent to participant: "+participant.getName()+" within choreography task "
			   +choreographyTask.getName()+" found.");	        
	   // log execution flow
	   LOGGER.exit();
	   return null;
   }

   public static String getMessageTypeNameChoreographyTaskSentFromParticipant(
		   ChoreographyTask choreographyTask, Participant participant){

	   // log execution flow
	   LOGGER.entry();	   
	   for (MessageFlow messageFlow : choreographyTask.getMessageFlowRef()) {
		   if(messageFlow.getSourceRef().equals(participant)){
			   LOGGER.info("The message type name of the message sent from participant: "+participant
					   .getName()+" within choreography task "+choreographyTask.getName()+" is "
					   +BPMNUtils.getMessageTypeName(messageFlow.getMessageRef()));			   
			   // log execution flow
			   LOGGER.exit();
			   return BPMNUtils.getMessageTypeName(messageFlow.getMessageRef());
		   }
	   }		
	   LOGGER.info("No message type name of message sent from participant: "+participant.getName()
	   		+" within choreography task "+choreographyTask.getName()+" found.");	        
	   // log execution flow
	   LOGGER.exit();
	   return null;
   }   
   
   public static String getMessageTypeNameChoreographyTaskSentToParticipant(
		   ChoreographyTask choreographyTask, Participant participant){

	   // log execution flow
	   LOGGER.entry();	 	   
	   for (MessageFlow messageFlow : choreographyTask.getMessageFlowRef()) {
		   if(messageFlow.getTargetRef().equals(participant)){
			   LOGGER.info("The message type name of the message sent to participant: "+participant.getName()
		   		+" within choreography task "+choreographyTask.getName()+" is "+BPMNUtils
		   		.getMessageTypeName(messageFlow.getMessageRef()));			   
			   // log execution flow
			   LOGGER.exit();
			   return BPMNUtils.getMessageTypeName(messageFlow.getMessageRef());
		   }
	   }		
	   LOGGER.info("No message type name of message sent to participant: "+participant.getName()
  		+" within choreography task "+choreographyTask.getName()+" found.");
	   // log execution flow
	   LOGGER.exit();
	   return null;
   }   
   
   public static ChoreographyMessageInformation getDataTypeChoreographyInformation(
		   List<Choreography> choreographies, String choreographyMessageName){

	   // log execution flow
	   LOGGER.entry();
	   ChoreographyMessageInformation choreographyMessageInformation = new ChoreographyMessageInformation();
	   choreographyMessageInformation.setChoreographyMessageName(choreographyMessageName);	   
	   boolean messageFound = false;
	   // iterate over every choreography	   
	   for (Iterator<Choreography> iterator = choreographies.iterator();
			   iterator.hasNext() && !messageFound;){
		   // get the choregraphy
		   Choreography choreography = (Choreography) iterator.next();
		   for (Iterator<FlowElement> iterator2 = choreography.getFlowElements().iterator();
				   iterator2.hasNext() && !messageFound;){
			   // get the flow element
			   FlowElement flowElement = (FlowElement) iterator2.next();
			   // check if flowElement is a choreography task
			   if(flowElement instanceof ChoreographyTask){
				   // flowElement is a choreography task
				   // iterate over task messages
				   for (MessageFlow messageFlow : ((ChoreographyTask) flowElement).getMessageFlowRef()){
					   // check if the message is the searched one
					   if(messageFlow.getMessageRef().getName().equals(choreographyMessageName)){
						   // the message is the searched one
						   // set choreographyMessageInformation
						   choreographyMessageInformation.setChoreographyTaskName(flowElement.getName()); 
						   choreographyMessageInformation.setInitiatingParticipantName(BPMNUtils
								   .getInitiatingParticipant((ChoreographyActivity) flowElement).getName());
						   choreographyMessageInformation.setReceivingParticipantName(BPMNUtils
								   .getReceivingParticipant((ChoreographyActivity) flowElement).getName());						   
						   choreographyMessageInformation.setDataTypeName(BPMNUtils
								   .getMessageTypeName(messageFlow.getMessageRef()));						   
						   choreographyMessageInformation
						   .setSenderParticipantName(((Participant) messageFlow.getSourceRef()).getName());
						   choreographyMessageInformation
						   .setReceiverParticipantName(((Participant) messageFlow.getTargetRef()).getName());
						   messageFound = true;
					   }
				   }				   
			   }			
		   }
	   }	 
	   // log execution flow
	   LOGGER.exit();
	   return choreographyMessageInformation;
   }
   
   public static String getMessageTypeName(Message message){
	   
	   return StringUtils.split(((org.eclipse.emf.ecore.impl.BasicEObjectImpl)message.getItemRef()
			   .getStructureRef()).eProxyURI().toString(),
			   EPROXY_URI_MESSAGE_ITEM_REF_STRUCTURE_REF_MESSAGE_TYPE_NAME_SEPARATOR)[1];	   
   }
   
   public static String getDataTypeNameConditionExpressionOfSequenceFlow(SequenceFlow sequenceFlow){
	   
	   return StringUtils.split(((org.eclipse.emf.ecore.impl.BasicEObjectImpl)(
				(FormalExpression) sequenceFlow.getConditionExpression())
				.getEvaluatesToTypeRef().getStructureRef()).eProxyURI().toString(), 
			   EPROXY_URI_MESSAGE_ITEM_REF_STRUCTURE_REF_MESSAGE_TYPE_NAME_SEPARATOR)[1];
   }
   
   public static String getConditionBodyConditionExpressionOfSequenceFlow(SequenceFlow sequenceFlow){
	
	   return ((FormalExpressionImpl) sequenceFlow.getConditionExpression()).getBody();
   }
   
   public static boolean isTaskInitiatingOfCorrelationData(List<TaskCorrelationData> tasksCorrelationData, 
		   ChoreographyTask task){
	   
	   // log execution flow
	   LOGGER.entry();   
	   for (TaskCorrelationData taskCorrelationData : tasksCorrelationData) {
		   if(taskCorrelationData.getInitiatingTaskName().equals(task.getName())){
			   // log execution flow
			   LOGGER.exit();
			   return true;
		   }	   
	   }
	   // log execution flow
	   LOGGER.exit();
	   return false;
   }
   
   public static boolean isTaskReceivingOfCorrelationData(List<TaskCorrelationData> tasksCorrelationData, 
		   ChoreographyTask task){

	   // log execution flow
	   LOGGER.entry(); 	   
	   for (TaskCorrelationData taskCorrelationData : tasksCorrelationData) {
		   if(taskCorrelationData.getReceivingTaskName().equals(task.getName())){
			   // log execution flow
			   LOGGER.exit();
			   return true;
		   }
	   }
	   // log execution flow
	   LOGGER.exit();
	   return false;
   } 
   
   public static String getInitiatingTaskNameCorrelatedToReceivingTask(
		   List<TaskCorrelationData> tasksCorrelationData, ChoreographyTask choreographyTask){
	
	   // log execution flow
	   LOGGER.entry(); 
	   for (TaskCorrelationData taskCorrelationData : tasksCorrelationData) {
		   if(taskCorrelationData.getReceivingTaskName().equals(choreographyTask.getName())){
			   LOGGER.info("The initiating choreography task correlated to choreography task "
					   +choreographyTask.getName()+" is "+taskCorrelationData.getInitiatingTaskName());
			   // log execution flow
			   LOGGER.exit();
			   return taskCorrelationData.getInitiatingTaskName();
		   }   
	   }
	   LOGGER.info("No initiating choreography task correlated to choreography task "
			   +choreographyTask.getName()+" found.");	        
	   throw new CDGeneratorException("No initiating choreography task correlated to choreography task "
			   +choreographyTask.getName()+" found not found see log file for details.");
   }    
         
   public static boolean isParticipantInvolvedInReceivingCorrelatedTask(Choreography choreography, 
		   List<TaskCorrelationData> tasksCorrelationData,Participant participant){

	   // log execution flow
	   LOGGER.entry();
	   for (TaskCorrelationData taskCorrelationData : tasksCorrelationData) {
		   ChoreographyTask task = BPMNUtils.getChoreographyTaskFromName(choreography,taskCorrelationData
				   .getReceivingTaskName());
		   if(BPMNUtils.getReceivingParticipant(task).equals(participant)){
			   // log execution flow
			   LOGGER.exit();
			   return true;
		   }	   
	   }
	   // log execution flow
	   LOGGER.exit();
	   return false;	   
   }
   
   public static ChoreographyTask getChoreographyTaskFromName(Choreography choreography,
		   String choreographyTaskName){
	   
	   // log execution flow
	   LOGGER.entry();	   
	   for (FlowElement flowElement : choreography.getFlowElements()) {
		   if (flowElement instanceof ChoreographyTask) {
			   if(flowElement.getName().equals(choreographyTaskName)){
				   LOGGER.info("Choreography task "+choreographyTaskName+" found into choreography "
						   +choreography.getName());
				   // log execution flow
				   LOGGER.exit();					
				   return (ChoreographyTask) flowElement;
			   }
		   }
	   }
	   LOGGER.info("No Choreography task "+choreographyTaskName+" found into choreography "
			   +choreography.getName());	        
	   throw new CDGeneratorException("No Choreography task "+choreographyTaskName
			   +" found into choreography "+choreography.getName()+" see log file for details.");
   }
   
   public static String createChoreographyTaskParticipantsBaseName(ChoreographyTask choreographyTask){
	   
	   return StringUtils.deleteWhitespace(new StringBuilder(BPMNUtils
			   .getInitiatingParticipant(choreographyTask).getName())
			   .append(BPMNUtils.getReceivingParticipant(choreographyTask).getName()).toString());
   }
   
   public static String getMessageNameSentToParticipantFromTaskName(Choreography choreography, 
		   List<TaskCorrelationData> tasksCorrelationData, String initiatingTaskName,
		   Participant participant){
	   
	   // log execution flow
	   LOGGER.entry();	   
	   for (TaskCorrelationData taskCorrelationData : tasksCorrelationData) {
		   if(taskCorrelationData.getInitiatingTaskName().equals(initiatingTaskName)){
			   ChoreographyTask receivingTask = BPMNUtils.getChoreographyTaskFromName(choreography, 
					   taskCorrelationData.getReceivingTaskName());
			   for (MessageFlow messageFlow : receivingTask.getMessageFlowRef()) {
				   if(messageFlow.getTargetRef().equals(participant)){
					   LOGGER.info("The message sent from participant "+participant.getName()+" is "
							   +messageFlow.getMessageRef().getName());
					   // log execution flow
					   LOGGER.exit();						   
					   return messageFlow.getMessageRef().getName();
				   }
			   }
		   }
	   }	   
	   LOGGER.info("No message sent from participant "+participant.getName()+" found.");	        
	   throw new CDGeneratorException("No message sent from participant "+participant.getName()
	   		+" found see log file for details.");
   }
   
   public static String getMessageTypeNameSentToParticipantFromTaskName(Choreography choreography, 
		   List<TaskCorrelationData> tasksCorrelationData, String initiatingTaskName,
		   Participant participant){

	   // log execution flow
	   LOGGER.entry();		   
	   for (TaskCorrelationData taskCorrelationData : tasksCorrelationData) {
		   if(taskCorrelationData.getInitiatingTaskName().equals(initiatingTaskName)){
			   ChoreographyTask receivingTask = BPMNUtils.getChoreographyTaskFromName(choreography, 
					   taskCorrelationData.getReceivingTaskName());
			   for (MessageFlow messageFlow : receivingTask.getMessageFlowRef()) {
				   if(messageFlow.getTargetRef().equals(participant)){
					   LOGGER.info("The message type name of the message sent from participant "
							   +participant.getName()+" is "+BPMNUtils.getMessageTypeName(messageFlow
							   .getMessageRef()));
					   // log execution flow
					   LOGGER.exit();	
					   return BPMNUtils.getMessageTypeName(messageFlow.getMessageRef());
				   }
			   }
		   }
	   }	   
	   LOGGER.info("The message type name of the message sent from participant "
			   +participant.getName()+" found.");	        
	   throw new CDGeneratorException("The message type name of the message sent from participant "
			   +participant.getName()+" found see log file for details.");
   } 
   
   public static List<String> getMessagesNamesFromSequenceFlowConditionBody(
		   String sequenceFlowConditionBody){

	   // log execution flow
	   LOGGER.entry();
	   List<String> messagesNames = new ArrayList<>();   
	   // remove all the xPathOperators from sequenceFlowConditionBody
	   for (String xPathOperator : XPATH_OPERATORS) {
		   sequenceFlowConditionBody = StringUtils.remove(sequenceFlowConditionBody, xPathOperator);
	   }
	   // split sequenceFlowConditionBody by space
	   String[] sequenceFlowConditionBodyItems =  StringUtils.split(sequenceFlowConditionBody, " ");
	   // iterate over sequenceFlowConditionBodyItems
	   for (int i = 0; i < sequenceFlowConditionBodyItems.length; i++) {
		   // check if sequenceFlowConditionBodyItems[i] is not a space and it is not a number
		   if(!sequenceFlowConditionBodyItems[i].equals(" ") && 
				   !StringUtils.isNumeric(sequenceFlowConditionBodyItems[i]) &&
				   !StringUtils.startsWith(sequenceFlowConditionBodyItems[i], "'")){
			   // sequenceFlowConditionBodyItems[i] is not a space and it is not a number and 
			   // it doesn't start with ' add to messagesNames the first string resulting from the split 
			   // of sequenceFlowConditionBodyItems[i] by / character
			   messagesNames.add(sequenceFlowConditionBodyItems[i].split("/")[0]);
		   }
	   }
	   // log execution flow
	   LOGGER.exit();
	   return messagesNames;
   }
      
   public static boolean checkIfParticipantInvolvedInitiatingOutgoingTasksDivergingExclusiveGateway(
		   ExclusiveGateway exclusiveGateway,Participant participant){

	   // log execution flow
	   LOGGER.entry();
	   for (int i = 0; i < exclusiveGateway.getOutgoing().size(); i++) {
			// take the i-est outgoing of flowNode
			ChoreographyTask task = (ChoreographyTask) exclusiveGateway.getOutgoing().get(i).getTargetRef();
			if(BPMNUtils.getInitiatingParticipant(task).equals(participant)){
				// log execution flow
				LOGGER.exit(true);				
				return true;
			}	
	   }
	   // log execution flow
	   LOGGER.exit(false);	   
	   return false;
   }
   
   public static boolean checkIfAllOutgoingDivergingExclusiveGatewayAreChoreographyTask(
		   ExclusiveGateway exclusiveGateway){
	   
	   // log execution flow
	   LOGGER.entry();
	   for (int i = 0; i < exclusiveGateway.getOutgoing().size(); i++) {
			if(!(exclusiveGateway.getOutgoing().get(i).getTargetRef() instanceof ChoreographyTask)){
				// log execution flow
				LOGGER.exit(false);				
				return false;
			}	
	   }
	   // log execution flow
	   LOGGER.exit(true);
	   return true;
   }	    
   
   public static boolean exclusiveGatewayHasToBeImplemented(FlowNode flowNode, Participant participant){
	   	   
		// log execution flow
		LOGGER.entry("CD Participant name "+participant.getName(),"Flow Node name "+flowNode.getName());
	   
	   // iterate over outgoing of an exclusive gateway
	   for (int i = 0; i < flowNode.getOutgoing().size(); i++) {	
		   // if the i-est outgoing is a choreography task and the initiating participant is the 
		   // participant in input 
  		   if((flowNode.getOutgoing().get(i).getTargetRef() instanceof ChoreographyTask) &&
			   BPMNUtils.getInitiatingParticipant((ChoreographyActivity) flowNode.getOutgoing().get(i)
					   .getTargetRef()).equals(participant)){	   
  			   // log execution flow
  			   LOGGER.exit(true);			   
			   return true;
		   }
  		   // if the i-est outgoing is neither a choreography task nor a converging gateway
  		   // it can be either an exclusive diverging or a parallel diverging gateway  		   
  		   if(!(flowNode.getOutgoing().get(i).getTargetRef() instanceof ChoreographyTask) &&
  			   !(((Gateway) flowNode).getGatewayDirection().compareTo(GatewayDirection.CONVERGING)==0)){
  			   // recursive call to
  			   // check there is a task for which the participant in input is the initiating participant
	  		   boolean participantInitiatingOfATask = BPMNUtils.exclusiveGatewayHasToBeImplemented(flowNode
	  				   .getOutgoing().get(i).getTargetRef(), participant);
	  		   // a task for which, the participant in input is the initiating participant, has been found
	  		   if(participantInitiatingOfATask){  			   
	  			   // log execution flow
	  			   LOGGER.exit(true);  			   
	  			   return true;
	  		   }	   
  		   }
	   }
	   
	   // log execution flow
	   LOGGER.exit(false);
	   
	   // from the exclusive gateway in input    
	   // no task has the participant in input as initiating participant
	   return false;	   
   }
   
	public static List<FlowNode> getFlowNodes(FlowNode flow){
		
		List<FlowNode> nodes = new ArrayList<>();
		traverseChoreographyFlows(flow, nodes);
		return nodes;
	}
	    
    public static void traverseChoreographyFlows(FlowNode flow, List<FlowNode> nodes) {
    	
    	if(flow instanceof EndEvent || nodes.contains((FlowNode) flow))
    		return;    		
    	for (SequenceFlow outgoing : flow.getOutgoing()) {
			nodes.add(flow);
			traverseChoreographyFlows(outgoing.getTargetRef(), nodes);
		}
    }
    
    public static String getTargetNamespaceChoreography(Choreography choreography){
    	
    	return ((Definitions) choreography.eContainer()).getTargetNamespace();
    }
    
    public static String getTargetNamespaceFromChoreographyData(ChoreographyData choreographyData){
    	
		// get the container from the first choreography in order to retrive the target namespace
    	return ((Definitions) choreographyData.getChoreographies().get(0).eContainer()).getTargetNamespace();
    }
        
	private static Choreography loadBPMNModel(File bpmnFile) throws CDGeneratorException {
		
		URI bpmnURI = URI.createURI(bpmnFile.toURI().toString());
		Choreography choreography = null;

		// register the BPMN2ResourceFactory in Factory registry
		Resource.Factory.Registry reg = Resource.Factory.Registry.INSTANCE;
		reg.getExtensionToFactoryMap().put("bpmn", new Bpmn2ResourceFactoryImpl());
		reg.getExtensionToFactoryMap().put("bpmn2", new Bpmn2ResourceFactoryImpl());

		ResourceSet resourceSet = new ResourceSetImpl();
		Resource resource = resourceSet.createResource(bpmnURI);

		try {
			resource.load(null);
			EObject root = resource.getContents().get(0);
			Definitions definitions;

			if (root instanceof DocumentRoot) {
				definitions = ((DocumentRoot) root).getDefinitions();
			} else {
				definitions = (Definitions) root;
			}

			for (EObject definition : definitions.eContents()) {
				if (definition instanceof Choreography) {
					choreography = (Choreography) definition;
				}
			}
			
		} catch (IOException e) {
			LOGGER.error(e.getMessage(),e);
			throw new CDGeneratorException("Error to load the resource " + resource.getURI().toFileString());
		}
		if (choreography == null) {
			throw new CDGeneratorException("No choreography found in the model " + resource.getURI()
				.toFileString());
		}
		return choreography;
	}

	private static List<Choreography> loadChoreographiesFromBPMN2Model(File bpmnFile){

		// log execution flow
		LOGGER.entry();
		List<Choreography> choreographies = new ArrayList<Choreography>();		
		// register the BPMN2ResourceFactory in Factory registry
		Resource.Factory.Registry reg = Resource.Factory.Registry.INSTANCE;
		reg.getExtensionToFactoryMap().put("bpmn", new Bpmn2ResourceFactoryImpl());
		reg.getExtensionToFactoryMap().put("bpmn2", new Bpmn2ResourceFactoryImpl());		
		URI bpmnURI = URI.createURI(bpmnFile.toURI().toString());		
		// load the resource and resolve
		ResourceSet resourceSet = new ResourceSetImpl();
		Resource resource = resourceSet.createResource(bpmnURI);
		
		try {
			// load the resource
			resource.load(null);
			DocumentRoot documentRoot = (DocumentRoot) resource.getContents().get(0);			
	        // get all the choreographies of the BPMN2 Model 
	        for (EObject definition : documentRoot.getDefinitions().eContents()) {
	            if (definition instanceof Choreography) {
	                Choreography choreography = (Choreography) definition;
	                choreographies.add(choreography);
	            }
	        }
		} catch (IOException e) {
			LOGGER.error(e.getMessage(),e);
			throw new CDGeneratorException(
					"Exception into loadChoreographiesFromBPMN2Mode see log file for details ");
		}
		// log execution flow
		LOGGER.exit();		
		return choreographies;
	}

	public static String getLoopIndexName(SubChoreography s) {
		
		// log execution flow
		LOGGER.entry();		
		if (s.getExtensionValues().size() > 0) {
			ExtensionAttributeValue eav = s.getExtensionValues().get(0);
			for (FeatureMap.Entry aValue : eav.getValue()) {
				AnyTypeImpl ati = (AnyTypeImpl) aValue.getValue();
				if (((String) ati.getAnyAttribute().getValue(0)).trim().equals("loopindexname")) {
					AnyTypeImpl val = (AnyTypeImpl) ati.getAny().get(0).getValue();
					// log index name found
					LOGGER.info("Loop index name found!");	
					// log execution flow
					LOGGER.exit();
					return (String) val.getMixed().get(0).getValue();
				}
			}
		}
		// log index name found
		LOGGER.info("Loop index name not found!");	
		// log execution flow
		LOGGER.exit();
		return null;
	}

	public static String getLoopMaximumIterationsNumber(SubChoreography s) {
		
		// log execution flow
		LOGGER.entry();	
		if (s.getExtensionValues().size() > 0) {
			ExtensionAttributeValue eav = s.getExtensionValues().get(0);
			for (FeatureMap.Entry aValue : eav.getValue()) {
				AnyTypeImpl ati = (AnyTypeImpl) aValue.getValue();
				if (((String) ati.getAnyAttribute().getValue(0)).trim()
						.equals("loopmaximumiterationsnumber")) {
					AnyTypeImpl val = (AnyTypeImpl) ati.getAny().get(0).getValue();
					// log index name found
					LOGGER.info("Loop Maximum Iterations Number found!");	
					// log execution flow
					LOGGER.exit();
					return (String) val.getMixed().get(0).getValue();
				}
			}
		}
		// log index name found
		LOGGER.info("Loop Maximum Iterations Number not found!");	
		// log execution flow
		LOGGER.exit();		
		return null;
	}	
	
}
